@@ -1,20 +1,17 @@
-import torch, flair
 import logging
 import re
-
 from abc import abstractmethod, ABC
-
 from collections import Counter
 from collections import defaultdict
-
-from deprecated import deprecated
-from flair.file_utils import Tqdm
 from operator import itemgetter
+from typing import List, Dict, Union, Optional
 
+import flair
+import torch
 from torch.utils.data import Dataset
 from torch.utils.data.dataset import ConcatDataset, Subset
 
-from typing import List, Dict, Union, Callable, Optional
+from flair.file_utils import Tqdm
 
 log = logging.getLogger("flair")
 
@@ -28,12 +25,18 @@ class Dictionary:
         # init dictionaries
         self.item2idx: Dict[str, int] = {}
         self.idx2item: List[str] = []
-        self.multi_label: bool = False
-
+        self.add_unk = add_unk
         # in order to deal with unknown tokens, add <unk>
         if add_unk:
             self.add_item("<unk>")
 
+    def remove_item(self, item: str):
+
+        item = item.encode("utf-8")
+        if item in self.item2idx:
+            self.idx2item.remove(item)
+            del self.item2idx[item]
+
     def add_item(self, item: str) -> int:
         """
         add string - if already in dictionary returns its ID. if not in dictionary, it will get a new ID.
@@ -52,11 +55,15 @@ class Dictionary:
         :param item: string for which ID is requested
         :return: ID of string, otherwise 0
         """
-        item = item.encode("utf-8")
-        if item in self.item2idx.keys():
-            return self.item2idx[item]
-        else:
+        item_encoded = item.encode("utf-8")
+        if item_encoded in self.item2idx.keys():
+            return self.item2idx[item_encoded]
+        elif self.add_unk:
             return 0
+        else:
+            log.error(f"The string '{item}' is not in dictionary! Dictionary contains only: {self.get_items()}")
+            log.error("You can create a Dictionary that handles unknown items with an <unk>-key by setting add_unk = True in the construction.")
+            raise IndexError
 
     def get_idx_for_items(self, items: List[str]) -> List[int]:
         """
@@ -95,6 +102,12 @@ class Dictionary:
         with open(savefile, "wb") as f:
             mappings = {"idx2item": self.idx2item, "item2idx": self.item2idx}
             pickle.dump(mappings, f)
+            
+    def __setstate__(self, d):
+        self.__dict__ = d
+        if 'add_unk' not in self.__dict__.keys():
+            self.__dict__['add_unk'] = False
+        
 
     @classmethod
     def load_from_file(cls, filename: str):
@@ -139,9 +152,16 @@ class Label:
     """
 
     def __init__(self, value: str, score: float = 1.0):
+        self._value = value
+        self._score = score
+        super().__init__()
+
+    def set_value(self, value: str, score: float = 1.0):
         self.value = value
         self.score = score
-        super().__init__()
+
+    def spawn(self, value: str, score: float = 1.0):
+        return Label(value, score)
 
     @property
     def value(self):
@@ -162,10 +182,7 @@ class Label:
 
     @score.setter
     def score(self, score):
-        if 0.0 <= score <= 1.0:
-            self._score = score
-        else:
-            self._score = 1.0
+        self._score = score
 
     def to_dict(self):
         return {"value": self.value, "confidence": self.score}
@@ -176,6 +193,67 @@ class Label:
     def __repr__(self):
         return f"{self._value} ({round(self._score, 4)})"
 
+    def __eq__(self, other):
+        return self.value == other.value and self.score == other.score
+
+    @property
+    def identifier(self):
+        return ""
+
+
+class SpanLabel(Label):
+    def __init__(self, span, value: str, score: float = 1.0):
+        super().__init__(value, score)
+        self.span = span
+
+    def spawn(self, value: str, score: float = 1.0):
+        return SpanLabel(self.span, value, score)
+
+    def __str__(self):
+        return f"{self._value} [{self.span.id_text}] ({round(self._score, 4)})"
+
+    def __repr__(self):
+        return f"{self._value} [{self.span.id_text}] ({round(self._score, 4)})"
+
+    def __len__(self):
+        return len(self.span)
+
+    def __eq__(self, other):
+        return self.value == other.value and self.score == other.score and self.span.id_text == other.span.id_text
+
+    @property
+    def identifier(self):
+        return f"{self.span.id_text}"
+
+
+class RelationLabel(Label):
+    def __init__(self, head, tail, value: str, score: float = 1.0):
+        super().__init__(value, score)
+        self.head = head
+        self.tail = tail
+
+    def spawn(self, value: str, score: float = 1.0):
+        return RelationLabel(self.head, self.tail, value, score)
+
+    def __str__(self):
+        return f"{self._value} [{self.head.id_text} -> {self.tail.id_text}] ({round(self._score, 4)})"
+
+    def __repr__(self):
+        return f"{self._value} from {self.head.id_text} -> {self.tail.id_text} ({round(self._score, 4)})"
+
+    def __len__(self):
+        return len(self.head) + len(self.tail)
+
+    def __eq__(self, other):
+        return self.value == other.value \
+               and self.score == other.score \
+               and self.head.id_text == other.head.id_text \
+               and self.tail.id_text == other.tail.id_text
+
+    @property
+    def identifier(self):
+        return f"{self.head.id_text} -> {self.tail.id_text}"
+
 
 class DataPoint:
     """
@@ -201,29 +279,40 @@ class DataPoint:
     def clear_embeddings(self, embedding_names: List[str] = None):
         pass
 
-    def add_label(self, label_type: str, value: str, score: float = 1.):
+    def add_label(self, typename: str, value: str, score: float = 1.):
 
-        if label_type not in self.annotation_layers:
-            self.annotation_layers[label_type] = [Label(value, score)]
+        if typename not in self.annotation_layers:
+            self.annotation_layers[typename] = [Label(value, score)]
         else:
-            self.annotation_layers[label_type].append(Label(value, score))
+            self.annotation_layers[typename].append(Label(value, score))
 
         return self
 
-    def set_label(self, label_type: str, value: str, score: float = 1.):
-        self.annotation_layers[label_type] = [Label(value, score)]
+    def add_complex_label(self, typename: str, label: Label):
+
+        if typename in self.annotation_layers and label in self.annotation_layers[typename]:
+            return self
+
+        if typename not in self.annotation_layers:
+            self.annotation_layers[typename] = [label]
+        else:
+            self.annotation_layers[typename].append(label)
 
         return self
 
-    def remove_labels(self, label_type: str):
-        if label_type in self.annotation_layers.keys():
-            del self.annotation_layers[label_type]
+    def set_label(self, typename: str, value: str, score: float = 1.):
+        self.annotation_layers[typename] = [Label(value, score)]
+        return self
 
-    def get_labels(self, label_type: str = None):
-        if label_type is None:
+    def remove_labels(self, typename: str):
+        if typename in self.annotation_layers.keys():
+            del self.annotation_layers[typename]
+
+    def get_labels(self, typename: str = None):
+        if typename is None:
             return self.labels
 
-        return self.annotation_layers[label_type] if label_type in self.annotation_layers else []
+        return self.annotation_layers[typename] if typename in self.annotation_layers else []
 
     @property
     def labels(self) -> List[Label]:
@@ -257,6 +346,9 @@ class DataPair(DataPoint):
     def to_plain_string(self):
         return f"DataPair: First {self.first}  ||  Second {self.second}"
 
+    def to_original_text(self):
+        return f"{self.first.to_original_text()} || {self.second.to_original_text()}"
+
     def __len__(self):
         return len(self.first) + len(self.second)
 
@@ -419,6 +511,14 @@ class Span(DataPoint):
 
         return str
 
+    def to_plain_string(self):
+        plain = ""
+        for token in self.tokens:
+            plain += token.text
+            if token.whitespace_after:
+                plain += " "
+        return plain.rstrip()
+
     def to_dict(self):
         return {
             "text": self.to_original_text(),
@@ -430,16 +530,20 @@ class Span(DataPoint):
     def __str__(self) -> str:
         ids = ",".join([str(t.idx) for t in self.tokens])
         label_string = " ".join([str(label) for label in self.labels])
-        labels = f'   [− Labels: {label_string}]' if self.labels is not None else ""
+        labels = f'   [− Labels: {label_string}]' if self.labels else ""
         return (
             'Span [{}]: "{}"{}'.format(ids, self.text, labels)
         )
 
+    @property
+    def id_text(self) -> str:
+        return f"{' '.join([t.text for t in self.tokens])} ({','.join([str(t.idx) for t in self.tokens])})"
+
     def __repr__(self) -> str:
         ids = ",".join([str(t.idx) for t in self.tokens])
         return (
             '<{}-span ({}): "{}">'.format(self.tag, ids, self.text)
-            if self.tag is not None
+            if len(self.labels) > 0
             else '<span ({}): "{}">'.format(ids, self.text)
         )
 
@@ -460,6 +564,10 @@ class Span(DataPoint):
     def score(self):
         return self.labels[0].score
 
+    @property
+    def position_string(self):
+        return '-'.join([str(token.idx) for token in self])
+
 
 class Tokenizer(ABC):
     r"""An abstract class representing a :class:`Tokenizer`.
@@ -480,59 +588,21 @@ class Tokenizer(ABC):
         return self.__class__.__name__
 
 
-@deprecated(version="0.5", reason="Use 'flair.tokenization.SpaceTokenizer' instead.")
-def space_tokenizer(text: str) -> List[Token]:
-    # We don't want to create a SpaceTokenizer object each time this function is called,
-    # so delegate the call directly to the static run_tokenize method
-    from flair.tokenization import SpaceTokenizer
-    return SpaceTokenizer.run_tokenize(text)
-
-
-@deprecated(version="0.5", reason="Use 'flair.tokenization.SegtokTokenizer' instead.")
-def segtok_tokenizer(text: str) -> List[Token]:
-    # We don't want to create a SegtokTokenizer object each time this function is called,
-    # so delegate the call directly to the static run_tokenize method
-    from flair.tokenization import SegtokTokenizer
-    return SegtokTokenizer.run_tokenize(text)
-
-
-@deprecated(version="0.5", reason="Use 'flair.tokenization.SpacyTokenizer' instead.")
-def build_spacy_tokenizer(model) -> Callable[[str], List[Token]]:
-    from flair.tokenization import SpacyTokenizer
-    spacy_tokenizer = SpacyTokenizer(model)
-
-    def tokenizer(text: str) -> List[Token]:
-        return spacy_tokenizer.tokenize(text)
-
-    return tokenizer
-
-
-@deprecated(version="0.5", reason="Use 'flair.tokenization.JapaneseTokenizer' instead.")
-def build_japanese_tokenizer(tokenizer: str = "MeCab"):
-    from flair.tokenization import JapaneseTokenizer
-    japanese_tokenizer = JapaneseTokenizer(tokenizer)
-
-    def tokenizer(text: str) -> List[Token]:
-        return japanese_tokenizer.tokenize(text)
-
-    return tokenizer
-
-
 class Sentence(DataPoint):
     """
        A Sentence is a list of tokens and is used to represent a sentence or text fragment.
     """
 
     def __init__(
-        self,
-        text: str = None,
-        use_tokenizer: Union[bool, Tokenizer] = True,
-        language_code: str = None,
-        start_position: int = None
+            self,
+            text: Union[str, List[str]] = None,
+            use_tokenizer: Union[bool, Tokenizer] = True,
+            language_code: str = None,
+            start_position: int = None
     ):
         """
         Class to hold all meta related to a text (tokens, predictions, language code, ...)
-        :param text: original string
+        :param text: original string (sentence), or a list of string tokens (words)
         :param use_tokenizer: a custom tokenizer (default is :class:`SpaceTokenizer`)
             more advanced options are :class:`SegTokTokenizer` to use segtok or :class:`SpacyTokenizer`
             to use Spacy library if available). Check the implementations of abstract class Tokenizer or
@@ -568,8 +638,12 @@ class Sentence(DataPoint):
 
         # if text is passed, instantiate sentence with tokens (words)
         if text is not None:
-            text = self._restore_windows_1252_characters(text)
-            [self.add_token(token) for token in tokenizer.tokenize(text)]
+            if isinstance(text, (list, tuple)):
+                [self.add_token(self._restore_windows_1252_characters(token))
+                 for token in text]
+            else:
+                text = self._restore_windows_1252_characters(text)
+                [self.add_token(token) for token in tokenizer.tokenize(text)]
 
         # log a warning if the dataset is empty
         if text == "":
@@ -579,6 +653,9 @@ class Sentence(DataPoint):
 
         self.tokenized = None
 
+        # some sentences represent a document boundary (but most do not)
+        self.is_document_boundary: bool = False
+
     def get_token(self, token_id: int) -> Token:
         for token in self.tokens:
             if token.idx == token_id:
@@ -595,7 +672,7 @@ class Sentence(DataPoint):
         token.text = token.text.replace('\ufeff', '')
 
         # data with zero-width characters cannot be handled
-        if token.text.strip() == '':
+        if token.text == '':
             return
 
         self.tokens.append(token)
@@ -624,7 +701,7 @@ class Sentence(DataPoint):
             tag_value = tag.value
 
             # non-set tags are OUT tags
-            if tag_value == "" or tag_value == "O":
+            if tag_value == "" or tag_value == "O" or tag_value == "_":
                 tag_value = "O-"
 
             # anything that is not a BIOES tag is a SINGLE tag
@@ -654,7 +731,7 @@ class Sentence(DataPoint):
                 if span_score > min_score:
                     span = Span(current_span)
                     span.add_label(
-                        label_type=label_type,
+                        typename=label_type,
                         value=sorted(tags.items(), key=lambda k_v: k_v[1], reverse=True)[0][0],
                         score=span_score)
                     spans.append(span)
@@ -676,7 +753,7 @@ class Sentence(DataPoint):
             if span_score > min_score:
                 span = Span(current_span)
                 span.add_label(
-                    label_type=label_type,
+                    typename=label_type,
                     value=sorted(tags.items(), key=lambda k_v: k_v[1], reverse=True)[0][0],
                     score=span_score)
                 spans.append(span)
@@ -936,6 +1013,58 @@ class Sentence(DataPoint):
 
         return re.sub(r"[\u0080-\u0099]", to_windows_1252, text)
 
+    def next_sentence(self):
+        """
+        Get the next sentence in the document (works only if context is set through dataloader or elsewhere)
+        :return: next Sentence in document if set, otherwise None
+        """
+        if '_next_sentence' in self.__dict__.keys():
+            return self._next_sentence
+
+        if '_position_in_dataset' in self.__dict__.keys():
+            dataset = self._position_in_dataset[0]
+            index = self._position_in_dataset[1] + 1
+            if index < len(dataset):
+                return dataset[index]
+
+        return None
+
+    def previous_sentence(self):
+        """
+        Get the previous sentence in the document (works only if context is set through dataloader or elsewhere)
+        :return: previous Sentence in document if set, otherwise None
+        """
+        if '_previous_sentence' in self.__dict__.keys():
+            return self._previous_sentence
+
+        if '_position_in_dataset' in self.__dict__.keys():
+            dataset = self._position_in_dataset[0]
+            index = self._position_in_dataset[1] - 1
+            if index >= 0:
+                return dataset[index]
+
+        return None
+
+    def is_context_set(self) -> bool:
+        """
+        Return True or False depending on whether context is set (for instance in dataloader or elsewhere)
+        :return: True if context is set, else False
+        """
+        return '_previous_sentence' in self.__dict__.keys() or '_position_in_dataset' in self.__dict__.keys()
+
+    def get_labels(self, label_type: str = None):
+
+        # TODO: crude hack - replace with something better
+        if label_type:
+            spans = self.get_spans(label_type)
+            for span in spans:
+                self.add_complex_label(label_type, label=SpanLabel(span, span.tag, span.score))
+
+        if label_type is None:
+            return self.labels
+
+        return self.annotation_layers[label_type] if label_type in self.annotation_layers else []
+
 
 class Image(DataPoint):
 
@@ -1007,12 +1136,13 @@ class Corpus:
             dev: FlairDataset = None,
             test: FlairDataset = None,
             name: str = "corpus",
+            sample_missing_splits: bool = True,
     ):
         # set name
         self.name: str = name
 
         # sample test data if none is provided
-        if test is None:
+        if test is None and sample_missing_splits:
             train_length = len(train)
             test_size: int = round(train_length / 10)
             splits = randomly_split_into_two_datasets(train, test_size)
@@ -1020,7 +1150,7 @@ class Corpus:
             train = splits[1]
 
         # sample dev data if none is provided
-        if dev is None:
+        if dev is None and sample_missing_splits:
             train_length = len(train)
             dev_size: int = round(train_length / 10)
             splits = randomly_split_into_two_datasets(train, dev_size)
@@ -1046,13 +1176,13 @@ class Corpus:
 
     def downsample(self, percentage: float = 0.1, downsample_train=True, downsample_dev=True, downsample_test=True):
 
-        if downsample_train:
+        if downsample_train and self._train:
             self._train = self._downsample_to_proportion(self.train, percentage)
 
-        if downsample_dev:
+        if downsample_dev and self._dev:
             self._dev = self._downsample_to_proportion(self.dev, percentage)
 
-        if downsample_test:
+        if downsample_test and self._test:
             self._test = self._downsample_to_proportion(self.test, percentage)
 
         return self
@@ -1094,7 +1224,6 @@ class Corpus:
 
         return subset
 
-
     @staticmethod
     def _filter_empty_sentences(dataset) -> Dataset:
 
@@ -1234,12 +1363,12 @@ class Corpus:
 
     def __str__(self) -> str:
         return "Corpus: %d train + %d dev + %d test sentences" % (
-            len(self.train),
-            len(self.dev),
-            len(self.test),
+            len(self.train) if self.train else 0,
+            len(self.dev) if self.dev else 0,
+            len(self.test) if self.test else 0,
         )
 
-    def make_label_dictionary(self, label_type: str = None) -> Dictionary:
+    def make_label_dictionary(self, label_type: str) -> Dictionary:
         """
         Creates a dictionary of all labels assigned to the sentences in the corpus.
         :return: dictionary of labels
@@ -1249,31 +1378,59 @@ class Corpus:
 
         from flair.datasets import DataLoader
 
-        data = ConcatDataset([self.train, self.test])
+        datasets = [self.train]
+
+        data = ConcatDataset(datasets)
+
         loader = DataLoader(data, batch_size=1)
 
         log.info("Computing label dictionary. Progress:")
+
+        # if there are token labels of provided type, use these. Otherwise use sentence labels
+        token_labels_exist = False
+
+        all_label_types = Counter()
+        all_sentence_labels = []
         for batch in Tqdm.tqdm(iter(loader)):
 
             for sentence in batch:
 
-                # check if sentence itself has labels
-                labels = sentence.get_labels(label_type) if label_type is not None else sentence.labels
+                # if we are looking for sentence-level labels
+                if not token_labels_exist:
+                    # check if sentence itself has labels
+                    labels = sentence.get_labels(label_type)
+                    all_label_types.update(sentence.annotation_layers.keys())
 
-                for label in labels:
-                    label_dictionary.add_item(label.value)
+                    for label in labels:
+                        if label.value not in all_sentence_labels: all_sentence_labels.append(label.value)
+
+                    if not label_dictionary.multi_label:
+                        if len(labels) > 1:
+                            label_dictionary.multi_label = True
 
                 # check for labels of words
                 if isinstance(sentence, Sentence):
                     for token in sentence.tokens:
+                        all_label_types.update(token.annotation_layers.keys())
                         for label in token.get_labels(label_type):
                             label_dictionary.add_item(label.value)
+                            token_labels_exist = True
 
-                if not label_dictionary.multi_label:
-                    if len(labels) > 1:
-                        label_dictionary.multi_label = True
+        # if this is not a token-level prediction problem, add sentence-level labels to dictionary
+        if not token_labels_exist:
+            for label in all_sentence_labels:
+                label_dictionary.add_item(label)
 
-        log.info(label_dictionary.idx2item)
+        if len(label_dictionary.idx2item) == 0:
+            log.error(
+                f"Corpus contains only the labels: {', '.join([f'{label[0]} (#{label[1]})' for label in all_label_types.most_common()])}")
+            log.error(f"You specified as label_type='{label_type}' which is not in this dataset!")
+
+            raise Exception
+
+        log.info(
+            f"Corpus contains the labels: {', '.join([label[0] + f' (#{label[1]})' for label in all_label_types.most_common()])}")
+        log.info(f"Created (for label '{label_type}') {label_dictionary}")
 
         return label_dictionary
 
@@ -1285,12 +1442,16 @@ class Corpus:
         return class_to_count
 
     def get_all_sentences(self) -> Dataset:
-        return ConcatDataset([self.train, self.dev, self.test])
+        parts = []
+        if self.train: parts.append(self.train)
+        if self.dev: parts.append(self.dev)
+        if self.test: parts.append(self.test)
+        return ConcatDataset(parts)
 
     def make_tag_dictionary(self, tag_type: str) -> Dictionary:
 
         # Make the tag dictionary
-        tag_dictionary: Dictionary = Dictionary()
+        tag_dictionary: Dictionary = Dictionary(add_unk=False)
         tag_dictionary.add_item("O")
         for sentence in self.get_all_sentences():
             for token in sentence.tokens:
@@ -1301,19 +1462,31 @@ class Corpus:
 
 
 class MultiCorpus(Corpus):
-    def __init__(self, corpora: List[Corpus], name: str = "multicorpus"):
+    def __init__(self, corpora: List[Corpus], name: str = "multicorpus", **corpusargs):
         self.corpora: List[Corpus] = corpora
 
+        train_parts = []
+        dev_parts = []
+        test_parts = []
+        for corpus in self.corpora:
+            if corpus.train: train_parts.append(corpus.train)
+            if corpus.dev: dev_parts.append(corpus.dev)
+            if corpus.test: test_parts.append(corpus.test)
+
         super(MultiCorpus, self).__init__(
-            ConcatDataset([corpus.train for corpus in self.corpora]),
-            ConcatDataset([corpus.dev for corpus in self.corpora]),
-            ConcatDataset([corpus.test for corpus in self.corpora]),
+            ConcatDataset(train_parts) if len(train_parts) > 0 else None,
+            ConcatDataset(dev_parts) if len(dev_parts) > 0 else None,
+            ConcatDataset(test_parts) if len(test_parts) > 0 else None,
             name=name,
+            **corpusargs,
         )
 
     def __str__(self):
-        output = f"MultiCorpus: {len(self.train)} train + {len(self.dev)} dev + {len(self.test)} test sentences\n - "
-        output += "\n - ".join([f'{type(corpus).__name__} {str(corpus)}' for corpus in self.corpora])
+        output = f"MultiCorpus: " \
+                 f"{len(self.train) if self.train else 0} train + " \
+                 f"{len(self.dev) if self.dev else 0} dev + " \
+                 f"{len(self.test) if self.test else 0} test sentences\n - "
+        output += "\n - ".join([f'{type(corpus).__name__} {str(corpus)} - {corpus.name}' for corpus in self.corpora])
         return output
 
 
@@ -1345,8 +1518,8 @@ def iob_iobes(tags):
     """
     new_tags = []
     for i, tag in enumerate(tags):
-        if tag.value == "O":
-            new_tags.append(tag.value)
+        if tag.value == "O" or tag.value == "":
+            new_tags.append("O")
         elif tag.value.split("-")[0] == "B":
             if i + 1 != len(tags) and tags[i + 1].value.split("-")[0] == "I":
                 new_tags.append(tag.value)
@@ -1361,8 +1534,8 @@ def iob_iobes(tags):
             raise Exception("Invalid IOB format!")
     return new_tags
 
-def randomly_split_into_two_datasets(dataset, length_of_first):
 
+def randomly_split_into_two_datasets(dataset, length_of_first):
     import random
     indices = [i for i in range(len(dataset))]
     random.shuffle(indices)
