@@ -1,21 +1,22 @@
-import os
-from typing import List, Dict, Union, Callable
-import re
 import logging
-from enum import Enum
-from pathlib import Path
+import os
+import re
 
 from deprecated import deprecated
+from enum import Enum
+from pathlib import Path
+from typing import List, Dict, Union
+
 
 import flair
 from flair.data import (
     Sentence,
     Corpus,
     Token,
-    MultiCorpus,
-    space_tokenizer,
-    segtok_tokenizer,
+    Tokenizer,
+    MultiCorpus
 )
+from flair.tokenization import SegtokTokenizer, SpaceTokenizer
 from flair.file_utils import cached_path
 
 log = logging.getLogger("flair")
@@ -119,15 +120,15 @@ class NLPTaskDataFetcher:
     @staticmethod
     @deprecated(version="0.4.1", reason="Use 'flair.datasets' instead.")
     def load_corpora(
-        tasks: List[Union[NLPTask, str]], base_path: Path = None
+        tasks: List[Union[NLPTask, str]], base_path: Union[str, Path] = None
     ) -> MultiCorpus:
         return MultiCorpus(
-            [NLPTaskDataFetcher.load_corpus(task, base_path) for task in tasks]
+            [NLPTaskDataFetcher.load_corpus(task, Path(base_path)) for task in tasks]
         )
 
     @staticmethod
     @deprecated(version="0.4.1", reason="Use 'flair.datasets' instead.")
-    def load_corpus(task: Union[NLPTask, str], base_path: [str, Path] = None) -> Corpus:
+    def load_corpus(task: Union[NLPTask, str], base_path: Union[str, Path] = None) -> Corpus:
         """
         Helper function to fetch a Corpus for a specific NLPTask. For this to work you need to first download
         and put into the appropriate folder structure the corresponding NLP task data. The tutorials on
@@ -144,7 +145,7 @@ class NLPTaskDataFetcher:
 
         # default dataset folder is the cache root
         if not base_path:
-            base_path = Path(flair.cache_root) / "datasets"
+            base_path = flair.cache_root / "datasets"
 
         if type(base_path) == str:
             base_path: Path = Path(base_path)
@@ -232,10 +233,10 @@ class NLPTaskDataFetcher:
             NLPTask.TREC_50.value,
             NLPTask.REGRESSION.value,
         ]:
-            tokenizer: Callable[[str], List[Token]] = space_tokenizer if task in [
+            tokenizer: Tokenizer = SpaceTokenizer() if task in [
                 NLPTask.TREC_6.value,
                 NLPTask.TREC_50.value,
-            ] else segtok_tokenizer
+            ] else SegtokTokenizer()
 
             return NLPTaskDataFetcher.load_classification_corpus(
                 data_folder, tokenizer=tokenizer
@@ -250,7 +251,7 @@ class NLPTaskDataFetcher:
 
         if task.startswith("wassa"):
             return NLPTaskDataFetcher.load_classification_corpus(
-                data_folder, tokenizer=segtok_tokenizer
+                data_folder, tokenizer=SegtokTokenizer()
             )
 
     @staticmethod
@@ -403,7 +404,7 @@ class NLPTaskDataFetcher:
         train_file=None,
         test_file=None,
         dev_file=None,
-        tokenizer: Callable[[str], List[Token]] = segtok_tokenizer,
+        tokenizer: Tokenizer = SegtokTokenizer(),
         max_tokens_per_doc=-1,
     ) -> Corpus:
         """
@@ -413,6 +414,7 @@ class NLPTaskDataFetcher:
         :param train_file: the name of the train file
         :param test_file: the name of the test file
         :param dev_file: the name of the dev file, if None, dev data is sampled from train
+        :param tokenizer: Custom tokenizer to use (default SegtokTokenizer)
         :return: a Corpus with annotated train, dev and test data
         """
 
@@ -477,7 +479,7 @@ class NLPTaskDataFetcher:
     def read_text_classification_file(
         path_to_file: Union[str, Path],
         max_tokens_per_doc=-1,
-        tokenizer=segtok_tokenizer,
+        tokenizer: Tokenizer = SegtokTokenizer(),
     ) -> List[Sentence]:
         """
         Reads a data file for text classification. The file should contain one document/text per line.
@@ -487,6 +489,7 @@ class NLPTaskDataFetcher:
         __label__<class_name_1> __label__<class_name_2> <text>
         :param path_to_file: the path to the data file
         :param max_tokens_per_doc: Takes at most this amount of tokens per document. If set to -1 all documents are taken as is.
+        :param tokenizer: Custom tokenizer to use to prepare the data set (default SegtokTokenizer)
         :return: list of sentences
         """
         label_prefix = "__label__"
@@ -521,7 +524,7 @@ class NLPTaskDataFetcher:
     @staticmethod
     @deprecated(version="0.4.1", reason="Use 'flair.datasets' instead.")
     def read_column_data(
-        path_to_column_file: Path,
+        path_to_column_file: Union[str, Path],
         column_name_map: Dict[int, str],
         infer_whitespace_after: bool = True,
     ):
@@ -570,7 +573,7 @@ class NLPTaskDataFetcher:
                 sentence: Sentence = Sentence()
 
             else:
-                fields: List[str] = re.split("\s+", line)
+                fields: List[str] = re.split(r"\s+", line)
                 token = Token(fields[text_column])
                 for column in column_name_map:
                     if len(fields) > column:
@@ -587,7 +590,7 @@ class NLPTaskDataFetcher:
 
     @staticmethod
     @deprecated(version="0.4.1", reason="Use 'flair.datasets' instead.")
-    def read_conll_ud(path_to_conll_file: Path) -> List[Sentence]:
+    def read_conll_ud(path_to_conll_file: Union[str, Path]) -> List[Sentence]:
         """
        Reads a file in CoNLL-U format and produces a list of Sentence with full morphosyntactic annotation
        :param path_to_conll_file: the path to the conll-u file
@@ -650,7 +653,7 @@ class NLPTaskDataFetcher:
         # conll 2000 chunking task
         if task == NLPTask.CONLL_2000:
             conll_2000_path = "https://www.clips.uantwerpen.be/conll2000/chunking/"
-            data_file = Path(flair.cache_root) / "datasets" / task.value / "train.txt"
+            data_file = flair.cache_root / "datasets" / task.value / "train.txt"
             if not data_file.is_file():
                 cached_path(
                     f"{conll_2000_path}train.txt.gz", Path("datasets") / task.value
@@ -661,27 +664,27 @@ class NLPTaskDataFetcher:
                 import gzip, shutil
 
                 with gzip.open(
-                    Path(flair.cache_root) / "datasets" / task.value / "train.txt.gz",
+                    flair.cache_root / "datasets" / task.value / "train.txt.gz",
                     "rb",
                 ) as f_in:
                     with open(
-                        Path(flair.cache_root) / "datasets" / task.value / "train.txt",
+                        flair.cache_root / "datasets" / task.value / "train.txt",
                         "wb",
                     ) as f_out:
                         shutil.copyfileobj(f_in, f_out)
                 with gzip.open(
-                    Path(flair.cache_root) / "datasets" / task.value / "test.txt.gz",
+                    flair.cache_root / "datasets" / task.value / "test.txt.gz",
                     "rb",
                 ) as f_in:
                     with open(
-                        Path(flair.cache_root) / "datasets" / task.value / "test.txt",
+                        flair.cache_root / "datasets" / task.value / "test.txt",
                         "wb",
                     ) as f_out:
                         shutil.copyfileobj(f_in, f_out)
 
         if task == NLPTask.NER_BASQUE:
             ner_basque_path = "http://ixa2.si.ehu.eus/eiec/"
-            data_path = Path(flair.cache_root) / "datasets" / task.value
+            data_path = flair.cache_root / "datasets" / task.value
             data_file = data_path / "named_ent_eu.train"
             if not data_file.is_file():
                 cached_path(
@@ -690,7 +693,7 @@ class NLPTaskDataFetcher:
                 import tarfile, shutil
 
                 with tarfile.open(
-                    Path(flair.cache_root) / "datasets" / task.value / "eiec_v1.0.tgz",
+                    flair.cache_root / "datasets" / task.value / "eiec_v1.0.tgz",
                     "r:gz",
                 ) as f_in:
                     corpus_files = (
@@ -705,14 +708,14 @@ class NLPTaskDataFetcher:
             imdb_acl_path = (
                 "http://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz"
             )
-            data_path = Path(flair.cache_root) / "datasets" / task.value
+            data_path = flair.cache_root / "datasets" / task.value
             data_file = data_path / "train.txt"
             if not data_file.is_file():
                 cached_path(imdb_acl_path, Path("datasets") / task.value)
                 import tarfile
 
                 with tarfile.open(
-                    Path(flair.cache_root)
+                    flair.cache_root
                     / "datasets"
                     / task.value
                     / "aclImdb_v1.tar.gz",
@@ -757,7 +760,7 @@ class NLPTaskDataFetcher:
                     Path("datasets") / task.value / "original",
                 )
 
-            data_path = Path(flair.cache_root) / "datasets" / task.value
+            data_path = flair.cache_root / "datasets" / task.value
             data_file = data_path / new_filenames[0]
 
             if not data_file.is_file():
@@ -824,7 +827,7 @@ class NLPTaskDataFetcher:
                 lc = "ru"
 
             data_file = (
-                Path(flair.cache_root)
+                flair.cache_root
                 / "datasets"
                 / task.value
                 / f"aij-wikiner-{lc}-wp3.train"
@@ -839,14 +842,14 @@ class NLPTaskDataFetcher:
 
                 # unpack and write out in CoNLL column-like format
                 bz_file = bz2.BZ2File(
-                    Path(flair.cache_root)
+                    flair.cache_root
                     / "datasets"
                     / task.value
                     / f"aij-wikiner-{lc}-wp3.bz2",
                     "rb",
                 )
                 with bz_file as f, open(
-                    Path(flair.cache_root)
+                    flair.cache_root
                     / "datasets"
                     / task.value
                     / f"aij-wikiner-{lc}-wp3.train",
@@ -1318,7 +1321,7 @@ class NLPTaskDataFetcher:
             for split in ["train", "dev", "test"]:
 
                 data_file = (
-                    Path(flair.cache_root)
+                    flair.cache_root
                     / "datasets"
                     / task.value
                     / f"{emotion}-{split}.txt"
@@ -1363,7 +1366,7 @@ class NLPTaskDataFetcher:
                 f"{ud_path}UD_German-HDT/dev/de_hdt-ud-train-b.conllu",
                 Path("datasets") / task.value / "original",
             )
-            data_path = Path(flair.cache_root) / "datasets" / task.value
+            data_path = flair.cache_root / "datasets" / task.value
 
             train_filenames = ["de_hdt-ud-train-a.conllu", "de_hdt-ud-train-b.conllu"]
 
