@@ -98,12 +98,15 @@ class WordEmbeddingsStore:
                 t = torch.tensor(self._get_vector(word=token.text.lower()))
                 token.set_embedding(self.name, t)
 
+    def get_names(self):
+        return [self.name]
+                
     @staticmethod
     def _get_store_path(embedding, backend='sqlite'):
         """
         get the filename of the store
         """
-        cache_dir = Path(flair.cache_root)
+        cache_dir = flair.cache_root
         embedding_filename = re.findall("/(embeddings/.*)", embedding.name)[0]
         store_path = cache_dir / (embedding_filename + "." + backend)
         return store_path
