@@ -4,6 +4,8 @@ from collections import Counter
 import torch
 from typing import List, Union
 
+from torch.nn.parameter import Parameter
+
 import flair
 from flair.data import Sentence
 from flair.embeddings import TokenEmbeddings, FlairEmbeddings, StackedEmbeddings, DocumentEmbeddings
@@ -318,6 +320,8 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
         self.rest_time = 0.
         self.batch_count = 0
 
+        self.to(flair.device)
+
     def train(self, mode=True):
         super().train(mode=mode)
         compress = False
@@ -361,7 +365,7 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
         for sentence in sentences:
             for token in sentence:
                 # add current embedding to the memory
-                local_embedding = token.get_subembedding(
+                local_embedding = token.get_embedding(
                     self.sub_embedding_names
                 )
 
@@ -390,6 +394,7 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
             dtype=torch.float,
             device=flair.device,
         )
+
         # default zero-state tensor
         zeros = torch.zeros(self.hidden_states, device=flair.device)
 
@@ -398,6 +403,8 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
 
         # prepare_stopwatch.start()
 
+        # print('all:')
+        # print(all_surface_form_tensor)
         # go through each unique surface form
         for s_id, surface_form in enumerate(surface_forms_sorted_by_memory_length):
 
@@ -414,8 +421,12 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
             #     surface_form_embedding_history.append(padding)
 
             for i in range(len(surface_form_embedding_history)):
+                # print('histry:')
+                # print(surface_form_embedding_history[i])
                 all_surface_form_tensor[i, s_id] = surface_form_embedding_history[i]
 
+            # print('all:')
+            # print(all_surface_form_tensor)
             # all_surface_form_tensor[s_id][: len(sentence)] = torch.stack(
             #     [token.get_embedding() for token in sentence], 0
             # )
@@ -460,7 +471,6 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
         # make batch tensors
         # all_surface_form_histories = torch.stack(all_surface_form_histories, 1)
         all_surface_form_histories = all_surface_form_tensor
-        # print(all_surface_form_histories)
         # print(all_surface_form_tensor)
         # asd
 
@@ -474,6 +484,7 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
 
         # dropout!
         all_surface_form_histories = self.dropout(all_surface_form_histories)
+        # print(all_surface_form_histories)
 
         # reproject if set
         if self.reproject:
@@ -561,6 +572,14 @@ class MemoryEmbeddingsFast(TokenEmbeddings):
     def embedding_length(self) -> int:
         return self.__embedding_length
 
+    def get_names(self) -> List[str]:
+        """Returns a list of embedding names. In most cases, it is just a list with one item, namely the name of
+        this embedding. But in some cases, the embedding is made up by different embeddings (StackedEmbedding).
+        Then, the list contains the names of all embeddings in the stack."""
+        names = [self.name]
+        names.extend(self.context_embeddings.get_names())
+        return names
+
 
 class MemoryEmbeddings(TokenEmbeddings):
     def __init__(
@@ -749,7 +768,7 @@ class MemoryEmbeddings(TokenEmbeddings):
             # go through each unique surface form and update state history
             for idx, surface_form in enumerate(surface_forms_sorted_by_memory_length):
                 # retrieve hidden states for this surface form, and the last initial hidden state
-                self.state_history[surface_form] = rnn_out[lengths[idx]-1, idx]
+                self.state_history[surface_form] = rnn_out[lengths[idx] - 1, idx]
 
         # print(self.state_history)
 
@@ -768,6 +787,14 @@ class MemoryEmbeddings(TokenEmbeddings):
 
                 embedding = rnn_out[memory_index, idx]
 
+                if self.concat_word_embeddings:
+                    import random
+                    stack = False if self.training and random.randint(1, 100) <= 50 else True
+
+                    embedding = torch.cat([embedding, token.get_embedding(self.sub_embedding_names)]) if stack \
+                        else torch.cat(
+                        [embedding, torch.zeros(self.embedding_length - self.hidden_states, device=flair.device)])
+
                 token.set_embedding(self.name, embedding)
 
         return sentences
@@ -777,4 +804,248 @@ class MemoryEmbeddings(TokenEmbeddings):
         return self.__embedding_length
 
     def extra_repr(self):
-        return f"[use_batch_memory_state='{self.use_batch_memory_state}', concat_word_embeddings='{self.concat_word_embeddings}']"
\ No newline at end of file
+        return f"[use_batch_memory_state='{self.use_batch_memory_state}', concat_word_embeddings='{self.concat_word_embeddings}']"
+
+    # def get_names(self) -> List[str]:
+    #     """Returns a list of embedding names. In most cases, it is just a list with one item, namely the name of
+    #     this embedding. But in some cases, the embedding is made up by different embeddings (StackedEmbedding).
+    #     Then, the list contains the names of all embeddings in the stack."""
+    #     names = [self.name]
+    #     names.extend(self.context_embeddings.get_names())
+    #     return names
+
+
+class OrderedMemoryEmbeddings(TokenEmbeddings):
+    def __init__(
+            self,
+            contextual_embeddings: TokenEmbeddings,
+            concat_word_embeddings: bool = True,
+            hidden_states: int = 64,
+            reproject: bool = False,
+            max_memory_length: int = 8,
+            dropout: float = 0.0,
+            use_batch_memory_state: bool = False,
+            train_initial_hidden_state: bool = True,
+    ):
+
+        super().__init__()
+
+        # variables
+        self.hidden_states = hidden_states
+        self.max_memory_length: int = max_memory_length
+        self.effective_memory_length: int = self.max_memory_length
+        self.concat_word_embeddings: bool = concat_word_embeddings
+        self.static_embeddings: bool = False
+        self.train_initial_hidden_state = train_initial_hidden_state
+
+        self.context_embeddings: TokenEmbeddings = contextual_embeddings
+
+        self.sub_embedding_names = (
+            [emb.name for emb in self.context_embeddings.embeddings]
+            if type(self.context_embeddings) is StackedEmbeddings
+            else [self.context_embeddings.name]
+        )
+
+        # determine embedding length
+        self.__embedding_length = (
+            self.context_embeddings.embedding_length + self.hidden_states
+            if self.concat_word_embeddings
+            else 0 + self.hidden_states
+        )
+
+        # the memory
+        self.last_instance = {}
+        self.last_state = {}
+
+        # the NN
+        self.dropout = torch.nn.Dropout(dropout)
+
+        self.reproject: bool = reproject
+        if self.reproject:
+            self.reprojection_layer = torch.nn.Linear(
+                self.context_embeddings.embedding_length,
+                self.context_embeddings.embedding_length,
+            )
+
+        self.rnn = torch.nn.GRU(
+            self.context_embeddings.embedding_length,
+            self.hidden_states,
+            num_layers=1,
+            bidirectional=False,
+        )
+        if self.train_initial_hidden_state:
+            self.initial_hidden = Parameter(torch.zeros(self.hidden_states), requires_grad=True)
+        else:
+            self.initial_hidden = torch.zeros(self.hidden_states)
+
+        self.name = self.context_embeddings.name + "+memory"
+
+        self.use_batch_memory_state = use_batch_memory_state
+
+        self.to(flair.device)
+
+    def train(self, mode=True):
+        super().train(mode=mode)
+
+        if mode:
+            # memory is wiped each time we do a training run
+            print("train mode resetting embeddings")
+            self.last_state = {}
+            self.effective_memory_length = self.max_memory_length
+            print(self.effective_memory_length)
+
+        else:
+            # memory is wiped each time we do a training run
+            print("prediction mode no backprop")
+            for word in self.last_instance.keys():
+                self.last_state[word] = getattr(self.last_instance[word], 'last_hidden')
+            self.effective_memory_length = 0
+
+            # print(self.last_state)
+
+    def _add_embeddings_internal(self, sentences: List[Sentence]) -> List[Sentence]:
+
+        word_history = {}
+
+        self.context_embeddings.embed(sentences)
+
+        # default zero-state tensor
+        # zeros = torch.zeros(self.hidden_states, device=flair.device)
+
+        # go through all tokens and set previous instance history if not yet set
+
+        if self.training:
+            idx = 0
+            for sentence in sentences:
+                for token in sentence:
+
+                    if not hasattr(token, 'previous_instance'):
+                        if token.text in self.last_instance.keys():
+                            setattr(token, 'previous_instance', self.last_instance[token.text])
+                        else:
+                            setattr(token, 'previous_instance', None)
+                    idx += 1
+
+                    self.last_instance[token.text] = token
+        else:
+            idx = 0
+            for sentence in sentences:
+                for token in sentence:
+                    idx += 1
+
+        # initialize zero-padded word embeddings tensor
+        all_surface_form_tensor = torch.zeros(
+            [
+                self.effective_memory_length + 1,
+                idx,
+                self.context_embeddings.embedding_length,
+            ],
+            dtype=torch.float,
+            device=flair.device,
+        )
+
+        all_hidden_tensor = torch.zeros(
+            [
+                1,
+                idx,
+                self.hidden_states,
+            ],
+            dtype=torch.float,
+            device=flair.device,
+        )
+
+        # print(all_surface_form_tensor.size())
+
+        idx = 0
+        lengths = []
+        for sentence in sentences:
+            for token in sentence:
+
+                word_history[idx] = [token]
+                memory = token
+                if self.training:
+                    while True:
+                        memory = memory.previous_instance
+                        if memory == None: break
+                        if len(word_history[idx]) >= self.effective_memory_length + 1: break
+
+                        word_history[idx].insert(0, memory)
+
+                for m_id, memory_token in enumerate(word_history[idx]):
+                    all_surface_form_tensor[m_id, idx] = memory_token.get_embedding(self.sub_embedding_names)
+
+                all_hidden_tensor[0, idx] = \
+                    self.initial_hidden if token.text not in self.last_state.keys() \
+                    else self.last_state[token.text]
+
+                lengths.append(len(word_history[idx]) -1) # length of history only (do not count current token)
+
+                idx += 1
+
+        # make batch tensors
+        all_surface_form_histories = all_surface_form_tensor
+
+        # dropout!
+        all_surface_form_histories = self.dropout(all_surface_form_histories)
+
+        # reproject if set
+        if self.reproject:
+            all_surface_form_histories = self.reprojection_layer(
+                all_surface_form_histories
+            )
+
+        # print(all_surface_form_histories.size())
+        rnn_out, hidden = self.rnn(all_surface_form_histories, all_hidden_tensor)
+        # print(rnn_out.size())
+
+        # finally, go through each token of each sentence and set the embedding
+        idx = 0
+        for sentence in sentences:
+
+            for token in sentence:
+
+                last_memory_index = lengths[idx] - 1
+                updated_memory_index = lengths[idx]
+
+                if last_memory_index >= 0:
+                    embedding = rnn_out[last_memory_index, idx]
+                else:
+                    embedding = self.initial_hidden if token.text not in self.last_state.keys() \
+                                else self.last_state[token.text]
+
+                # remember last hidden state of token
+                last_hidden = rnn_out[updated_memory_index, idx]
+
+                if self.training:
+                    setattr(token, 'last_hidden', last_hidden.clone().detach())
+                else:
+                    self.last_state[token.text] = last_hidden.clone().detach()
+
+                if self.concat_word_embeddings:
+                    import random
+                    stack = False if self.training and random.randint(1, 100) <= 50 else True
+
+                    embedding = torch.cat([embedding, token.get_embedding(self.sub_embedding_names)]) if stack \
+                        else torch.cat(
+                        [embedding, torch.zeros(self.embedding_length - self.hidden_states, device=flair.device)])
+
+                token.set_embedding(self.name, embedding)
+
+                idx += 1
+
+        return sentences
+
+    @property
+    def embedding_length(self) -> int:
+        return self.__embedding_length
+
+    def extra_repr(self):
+        return f"[use_batch_memory_state='{self.use_batch_memory_state}', concat_word_embeddings='{self.concat_word_embeddings}']"
+
+    def __getstate__(self):
+
+        for word in self.last_instance.keys():
+            self.last_state[word] = getattr(self.last_instance[word], 'last_hidden')
+        self.last_instance = {}
+
+        return self.__dict__
