@@ -2,16 +2,16 @@ import logging
 import sys
 
 from pathlib import Path
-from typing import List, Union, Optional, Dict
+from typing import List, Union, Optional, Dict, Tuple
 from warnings import warn
 
 import numpy as np
 import torch
 import torch.nn
 import torch.nn.functional as F
+from requests import HTTPError
 from tabulate import tabulate
 from torch.nn.parameter import Parameter
-from torch.utils.data.dataset import Dataset
 from tqdm import tqdm
 
 import flair.nn
@@ -19,7 +19,7 @@ from flair.data import Dictionary, Sentence, Label
 from flair.datasets import SentenceDataset, DataLoader
 from flair.embeddings import TokenEmbeddings, StackedEmbeddings, Embeddings
 from flair.file_utils import cached_path, unzip_file
-from flair.training_utils import Metric, Result, store_embeddings
+from flair.training_utils import store_embeddings
 
 log = logging.getLogger("flair")
 
@@ -65,7 +65,7 @@ def pad_tensors(tensor_list):
     return template, lens_
 
 
-class SequenceTagger(flair.nn.Model):
+class SequenceTagger(flair.nn.Classifier):
     def __init__(
             self,
             hidden_size: int,
@@ -81,7 +81,6 @@ class SequenceTagger(flair.nn.Model):
             reproject_embeddings: Union[bool, int] = True,
             train_initial_hidden_state: bool = False,
             rnn_type: str = "LSTM",
-            pickle_module: str = "pickle",
             beta: float = 1.0,
             loss_weights: Dict[str, float] = None,
     ):
@@ -105,8 +104,8 @@ class SequenceTagger(flair.nn.Model):
         (if any tag's weight is unspecified it will default to 1.0)
 
         """
-
         super(SequenceTagger, self).__init__()
+
         self.use_rnn = use_rnn
         self.hidden_size = hidden_size
         self.use_crf: bool = use_crf
@@ -149,8 +148,6 @@ class SequenceTagger(flair.nn.Model):
         self.use_word_dropout: float = word_dropout
         self.use_locked_dropout: float = locked_dropout
 
-        self.pickle_module = pickle_module
-
         if dropout > 0.0:
             self.dropout = torch.nn.Dropout(dropout)
 
@@ -213,7 +210,7 @@ class SequenceTagger(flair.nn.Model):
             )
         else:
             self.linear = torch.nn.Linear(
-                self.embeddings.embedding_length, len(tag_dictionary)
+                rnn_input_dim, len(tag_dictionary)
             )
 
         if self.use_crf:
@@ -353,6 +350,7 @@ class SequenceTagger(flair.nn.Model):
                 dataloader = tqdm(dataloader)
 
             overall_loss = 0
+            overall_count = 0
             batch_no = 0
             for batch in dataloader:
 
@@ -369,7 +367,9 @@ class SequenceTagger(flair.nn.Model):
                 feature = self.forward(batch)
 
                 if return_loss:
-                    overall_loss += self._calculate_loss(feature, batch)
+                    loss_and_count = self._calculate_loss(feature, batch)
+                    overall_loss += loss_and_count[0]
+                    overall_count += loss_and_count[1]
 
                 tags, all_tags = self._obtain_labels(
                     feature=feature,
@@ -391,211 +391,7 @@ class SequenceTagger(flair.nn.Model):
                 store_embeddings(batch, storage_mode=embedding_storage_mode)
 
             if return_loss:
-                return overall_loss / batch_no
-
-    def _requires_span_F1_evaluation(self) -> bool:
-        span_F1 = False
-        for item in self.tag_dictionary.get_items():
-            if item.startswith('B-'):
-                span_F1 = True
-        return span_F1
-
-    def _evaluate_with_span_F1(self, data_loader, embedding_storage_mode, mini_batch_size, out_path):
-        eval_loss = 0
-
-        batch_no: int = 0
-
-        metric = Metric("Evaluation", beta=self.beta)
-
-        lines: List[str] = []
-
-        y_true = []
-        y_pred = []
-
-        for batch in data_loader:
-
-            # predict for batch
-            loss = self.predict(batch,
-                                embedding_storage_mode=embedding_storage_mode,
-                                mini_batch_size=mini_batch_size,
-                                label_name='predicted',
-                                return_loss=True)
-            eval_loss += loss
-            batch_no += 1
-
-            for sentence in batch:
-
-                # make list of gold tags
-                gold_spans = sentence.get_spans(self.tag_type)
-                gold_tags = [(span.tag, repr(span)) for span in gold_spans]
-
-                # make list of predicted tags
-                predicted_spans = sentence.get_spans("predicted")
-                predicted_tags = [(span.tag, repr(span)) for span in predicted_spans]
-
-                # check for true positives, false positives and false negatives
-                for tag, prediction in predicted_tags:
-                    if (tag, prediction) in gold_tags:
-                        metric.add_tp(tag)
-                    else:
-                        metric.add_fp(tag)
-
-                for tag, gold in gold_tags:
-                    if (tag, gold) not in predicted_tags:
-                        metric.add_fn(tag)
-
-                tags_gold = []
-                tags_pred = []
-
-                # also write to file in BIO format to use old conlleval script
-                if out_path:
-                    for token in sentence:
-                        # check if in gold spans
-                        gold_tag = 'O'
-                        for span in gold_spans:
-                            if token in span:
-                                gold_tag = 'B-' + span.tag if token == span[0] else 'I-' + span.tag
-                        tags_gold.append(gold_tag)
-
-                        predicted_tag = 'O'
-                        # check if in predicted spans
-                        for span in predicted_spans:
-                            if token in span:
-                                predicted_tag = 'B-' + span.tag if token == span[0] else 'I-' + span.tag
-                        tags_pred.append(predicted_tag)
-
-                        lines.append(f'{token.text} {gold_tag} {predicted_tag}\n')
-                    lines.append('\n')
-
-                y_true.append(tags_gold)
-                y_pred.append(tags_pred)
-
-        if out_path:
-            with open(Path(out_path), "w", encoding="utf-8") as outfile:
-                outfile.write("".join(lines))
-
-        eval_loss /= batch_no
-
-        detailed_result = (
-            "\nResults:"
-            f"\n- F1-score (micro) {metric.micro_avg_f_score():.4f}"
-            f"\n- F1-score (macro) {metric.macro_avg_f_score():.4f}"
-            '\n\nBy class:'
-        )
-
-        for class_name in metric.get_classes():
-            detailed_result += (
-                f"\n{class_name:<10} tp: {metric.get_tp(class_name)} - fp: {metric.get_fp(class_name)} - "
-                f"fn: {metric.get_fn(class_name)} - precision: "
-                f"{metric.precision(class_name):.4f} - recall: {metric.recall(class_name):.4f} - "
-                f"f1-score: "
-                f"{metric.f_score(class_name):.4f}"
-            )
-
-        result = Result(
-            main_score=metric.micro_avg_f_score(),
-            log_line=f"{metric.precision():.4f}\t{metric.recall():.4f}\t{metric.micro_avg_f_score():.4f}",
-            log_header="PRECISION\tRECALL\tF1",
-            detailed_results=detailed_result,
-        )
-
-        return result, eval_loss
-
-    def evaluate(
-            self,
-            sentences: Union[List[Sentence], Dataset],
-            out_path: Union[str, Path] = None,
-            embedding_storage_mode: str = "none",
-            mini_batch_size: int = 32,
-            num_workers: int = 8,
-    ) -> (Result, float):
-
-        # read Dataset into data loader (if list of sentences passed, make Dataset first)
-        if not isinstance(sentences, Dataset):
-            sentences = SentenceDataset(sentences)
-        data_loader = DataLoader(sentences, batch_size=mini_batch_size, num_workers=num_workers)
-
-        # if span F1 needs to be used, use separate eval method
-        if self._requires_span_F1_evaluation():
-            return self._evaluate_with_span_F1(data_loader, embedding_storage_mode, mini_batch_size, out_path)
-
-        # else, use scikit-learn to evaluate
-        y_true = []
-        y_pred = []
-        labels = Dictionary(add_unk=False)
-
-        eval_loss = 0
-        batch_no: int = 0
-
-        lines: List[str] = []
-
-        for batch in data_loader:
-
-            # predict for batch
-            loss = self.predict(batch,
-                                embedding_storage_mode=embedding_storage_mode,
-                                mini_batch_size=mini_batch_size,
-                                label_name='predicted',
-                                return_loss=True)
-            eval_loss += loss
-            batch_no += 1
-
-            for sentence in batch:
-
-                for token in sentence:
-                    # add gold tag
-                    gold_tag = token.get_tag(self.tag_type).value
-                    y_true.append(labels.add_item(gold_tag))
-
-                    # add predicted tag
-                    predicted_tag = token.get_tag('predicted').value
-                    y_pred.append(labels.add_item(predicted_tag))
-
-                    # for file output
-                    lines.append(f'{token.text} {gold_tag} {predicted_tag}\n')
-
-                lines.append('\n')
-
-        if out_path:
-            with open(Path(out_path), "w", encoding="utf-8") as outfile:
-                outfile.write("".join(lines))
-
-        eval_loss /= batch_no
-
-        # use sklearn
-        from sklearn import metrics
-
-        # make "classification report"
-        target_names = []
-        for i in range(len(labels)):
-            target_names.append(labels.get_item_for_index(i))
-        classification_report = metrics.classification_report(y_true, y_pred, digits=4, target_names=target_names,
-                                                              zero_division=1)
-
-        # get scores
-        macro_f_score = round(metrics.fbeta_score(y_true, y_pred, beta=self.beta, average='micro'), 4)
-        micro_f_score = round(metrics.fbeta_score(y_true, y_pred, beta=self.beta, average='macro'), 4)
-        accuracy_score = round(metrics.accuracy_score(y_true, y_pred), 4)
-
-        detailed_result = (
-                "\nResults:"
-                f"\n- F-score (micro) {macro_f_score}"
-                f"\n- F-score (macro) {micro_f_score}"
-                f"\n- Accuracy {accuracy_score}"
-                '\n\nBy class:\n' + classification_report
-        )
-
-        # line for log file
-        log_header = "ACCURACY"
-        log_line = f"\t{accuracy_score}"
-
-        result = Result(
-            main_score=macro_f_score,
-            log_line=log_line,
-            log_header=log_header,
-            detailed_results=detailed_result,
-        )
-        return result, eval_loss
+                return overall_loss, overall_count
 
     def forward_loss(
             self, data_points: Union[List[Sentence], Sentence], sort=True
@@ -718,17 +514,19 @@ class SequenceTagger(flair.nn.Model):
 
     def _calculate_loss(
             self, features: torch.tensor, sentences: List[Sentence]
-    ) -> float:
+    ) -> Tuple[float, int]:
 
         lengths: List[int] = [len(sentence.tokens) for sentence in sentences]
 
         tag_list: List = []
+        token_count = 0
         for s_id, sentence in enumerate(sentences):
             # get the tags in this sentence
             tag_idx: List[int] = [
                 self.tag_dictionary.get_idx_for_item(token.get_tag(self.tag_type).value)
                 for token in sentence
             ]
+            token_count += len(tag_idx)
             # add tags as tensor
             tag = torch.tensor(tag_idx, device=flair.device)
             tag_list.append(tag)
@@ -742,7 +540,7 @@ class SequenceTagger(flair.nn.Model):
 
             score = forward_score - gold_score
 
-            return score.mean()
+            return score.sum(), token_count
 
         else:
             score = 0
@@ -751,10 +549,10 @@ class SequenceTagger(flair.nn.Model):
             ):
                 sentence_feats = sentence_feats[:sentence_length]
                 score += torch.nn.functional.cross_entropy(
-                    sentence_feats, sentence_tags, weight=self.loss_weights
+                    sentence_feats, sentence_tags, weight=self.loss_weights, reduction='sum',
                 )
-            score /= len(features)
-            return score
+
+            return score, token_count
 
     def _obtain_labels(
             self,
@@ -970,9 +768,52 @@ class SequenceTagger(flair.nn.Model):
     @staticmethod
     def _fetch_model(model_name) -> str:
 
+        # core Flair models on Huggingface ModelHub
+        huggingface_model_map = {
+            "ner": "flair/ner-english",
+            "ner-fast": "flair/ner-english-fast",
+            "ner-ontonotes": "flair/ner-english-ontonotes",
+            "ner-ontonotes-fast": "flair/ner-english-ontonotes-fast",
+            # Large NER models,
+            "ner-large": "flair/ner-english-large",
+            "ner-ontonotes-large": "flair/ner-english-ontonotes-large",
+            "de-ner-large": "flair/ner-german-large",
+            "nl-ner-large": "flair/ner-dutch-large",
+            "es-ner-large": "flair/ner-spanish-large",
+            # Multilingual NER models
+            "ner-multi": "flair/ner-multi",
+            "multi-ner": "flair/ner-multi",
+            "ner-multi-fast": "flair/ner-multi-fast",
+            # English POS models
+            "upos": "flair/upos-english",
+            "upos-fast": "flair/upos-english-fast",
+            "pos": "flair/pos-english",
+            "pos-fast": "flair/pos-english-fast",
+            # Multilingual POS models
+            "pos-multi": "flair/upos-multi",
+            "multi-pos": "flair/upos-multi",
+            "pos-multi-fast": "flair/upos-multi-fast",
+            "multi-pos-fast": "flair/upos-multi-fast",
+            # English SRL models
+            "frame": "flair/frame-english",
+            "frame-fast": "flair/frame-english-fast",
+            # English chunking models
+            "chunk": "flair/chunk-english",
+            "chunk-fast": "flair/chunk-english-fast",
+            # Language-specific NER models
+            "ar-ner": "megantosh/flair-arabic-multi-ner",
+            "ar-pos": "megantosh/flair-arabic-dialects-codeswitch-egy-lev",
+            "da-ner": "flair/ner-danish",
+            "de-ner": "flair/ner-german",
+            "de-ler": "flair/ner-german-legal",
+            "de-ner-legal": "flair/ner-german-legal",
+            "fr-ner": "flair/ner-french",
+            "nl-ner": "flair/ner-dutch",
+        }
+
         hu_path: str = "https://nlp.informatik.hu-berlin.de/resources/models"
 
-        model_map = {
+        hu_model_map = {
             # English NER models
             "ner": "/".join([hu_path, "ner", "en-ner-conll03-v0.4.pt"]),
             "ner-pooled": "/".join([hu_path, "ner-pooled", "en-ner-conll03-pooled-v0.5.pt"]),
@@ -985,7 +826,7 @@ class SequenceTagger(flair.nn.Model):
             "ner-multi-fast": "/".join([hu_path, "multi-ner-fast", "ner-multi-fast.pt"]),
             # English POS models
             "upos": "/".join([hu_path, "upos", "en-pos-ontonotes-v0.4.pt"]),
-            "upos-fast": "/".join([hu_path, "upos-fast", "en-pos-ontonotes-fast-v0.4.pt"]),
+            "upos-fast": "/".join([hu_path, "upos-fast", "en-upos-ontonotes-fast-v0.4.pt"]),
             "pos": "/".join([hu_path, "pos", "en-pos-ontonotes-v0.5.pt"]),
             "pos-fast": "/".join([hu_path, "pos-fast", "en-pos-ontonotes-fast-v0.5.pt"]),
             # Multilingual POS models
@@ -1007,10 +848,12 @@ class SequenceTagger(flair.nn.Model):
             "de-pos-tweets": "/".join([hu_path, "de-pos-tweets", "de-pos-twitter-v0.1.pt"]),
             "de-ner": "/".join([hu_path, "de-ner", "de-ner-conll03-v0.4.pt"]),
             "de-ner-germeval": "/".join([hu_path, "de-ner-germeval", "de-ner-germeval-0.4.1.pt"]),
+            "de-ler": "/".join([hu_path, "de-ner-legal", "de-ner-legal.pt"]),
+            "de-ner-legal": "/".join([hu_path, "de-ner-legal", "de-ner-legal.pt"]),
             # French models
             "fr-ner": "/".join([hu_path, "fr-ner", "fr-ner-wikiner-0.4.pt"]),
             # Dutch models
-            "nl-ner": "/".join([hu_path, "nl-ner", "nl-ner-bert-conll02-v0.5b.pt"]),
+            "nl-ner": "/".join([hu_path, "nl-ner", "nl-ner-bert-conll02-v0.8.pt"]),
             "nl-ner-rnn": "/".join([hu_path, "nl-ner-rnn", "nl-ner-conll02-v0.5.pt"]),
             # Malayalam models
             "ml-pos": "https://raw.githubusercontent.com/qburst/models-repository/master/FlairMalayalamModels/malayalam-xpos-model.pt",
@@ -1054,39 +897,107 @@ class SequenceTagger(flair.nn.Model):
             )}
 
         cache_dir = Path("models")
-        if model_name in model_map:
-            model_name = cached_path(model_map[model_name], cache_dir=cache_dir)
 
-        # the historical German taggers by the @redewiegergabe project
-        if model_name == "de-historic-indirect":
-            model_file = Path(flair.cache_root) / cache_dir / 'indirect' / 'final-model.pt'
+        get_from_model_hub = False
+
+        # check if model name is a valid local file
+        if Path(model_name).exists():
+            model_path = model_name
+
+        # check if model key is remapped to HF key - if so, print out information
+        elif model_name in huggingface_model_map:
+
+            # get mapped name
+            hf_model_name = huggingface_model_map[model_name]
+
+            # output information
+            log.info("-" * 80)
+            log.info(
+                f"The model key '{model_name}' now maps to 'https://huggingface.co/{hf_model_name}' on the HuggingFace ModelHub")
+            log.info(f" - The most current version of the model is automatically downloaded from there.")
+            if model_name in hu_model_map:
+                log.info(
+                    f" - (you can alternatively manually download the original model at {hu_model_map[model_name]})")
+            log.info("-" * 80)
+
+            # use mapped name instead
+            model_name = hf_model_name
+            get_from_model_hub = True
+
+        # if not, check if model key is remapped to direct download location. If so, download model
+        elif model_name in hu_model_map:
+            model_path = cached_path(hu_model_map[model_name], cache_dir=cache_dir)
+
+        # special handling for the taggers by the @redewiegergabe project (TODO: move to model hub)
+        elif model_name == "de-historic-indirect":
+            model_file = flair.cache_root / cache_dir / 'indirect' / 'final-model.pt'
             if not model_file.exists():
                 cached_path('http://www.redewiedergabe.de/models/indirect.zip', cache_dir=cache_dir)
-                unzip_file(Path(flair.cache_root) / cache_dir / 'indirect.zip', Path(flair.cache_root) / cache_dir)
-            model_name = str(Path(flair.cache_root) / cache_dir / 'indirect' / 'final-model.pt')
+                unzip_file(flair.cache_root / cache_dir / 'indirect.zip', flair.cache_root / cache_dir)
+            model_path = str(flair.cache_root / cache_dir / 'indirect' / 'final-model.pt')
 
-        if model_name == "de-historic-direct":
-            model_file = Path(flair.cache_root) / cache_dir / 'direct' / 'final-model.pt'
+        elif model_name == "de-historic-direct":
+            model_file = flair.cache_root / cache_dir / 'direct' / 'final-model.pt'
             if not model_file.exists():
                 cached_path('http://www.redewiedergabe.de/models/direct.zip', cache_dir=cache_dir)
-                unzip_file(Path(flair.cache_root) / cache_dir / 'direct.zip', Path(flair.cache_root) / cache_dir)
-            model_name = str(Path(flair.cache_root) / cache_dir / 'direct' / 'final-model.pt')
+                unzip_file(flair.cache_root / cache_dir / 'direct.zip', flair.cache_root / cache_dir)
+            model_path = str(flair.cache_root / cache_dir / 'direct' / 'final-model.pt')
 
-        if model_name == "de-historic-reported":
-            model_file = Path(flair.cache_root) / cache_dir / 'reported' / 'final-model.pt'
+        elif model_name == "de-historic-reported":
+            model_file = flair.cache_root / cache_dir / 'reported' / 'final-model.pt'
             if not model_file.exists():
                 cached_path('http://www.redewiedergabe.de/models/reported.zip', cache_dir=cache_dir)
-                unzip_file(Path(flair.cache_root) / cache_dir / 'reported.zip', Path(flair.cache_root) / cache_dir)
-            model_name = str(Path(flair.cache_root) / cache_dir / 'reported' / 'final-model.pt')
+                unzip_file(flair.cache_root / cache_dir / 'reported.zip', flair.cache_root / cache_dir)
+            model_path = str(flair.cache_root / cache_dir / 'reported' / 'final-model.pt')
 
-        if model_name == "de-historic-free-indirect":
-            model_file = Path(flair.cache_root) / cache_dir / 'freeIndirect' / 'final-model.pt'
+        elif model_name == "de-historic-free-indirect":
+            model_file = flair.cache_root / cache_dir / 'freeIndirect' / 'final-model.pt'
             if not model_file.exists():
                 cached_path('http://www.redewiedergabe.de/models/freeIndirect.zip', cache_dir=cache_dir)
-                unzip_file(Path(flair.cache_root) / cache_dir / 'freeIndirect.zip', Path(flair.cache_root) / cache_dir)
-            model_name = str(Path(flair.cache_root) / cache_dir / 'freeIndirect' / 'final-model.pt')
+                unzip_file(flair.cache_root / cache_dir / 'freeIndirect.zip', flair.cache_root / cache_dir)
+            model_path = str(flair.cache_root / cache_dir / 'freeIndirect' / 'final-model.pt')
 
-        return model_name
+        # for all other cases (not local file or special download location), use HF model hub
+        else:
+            get_from_model_hub = True
+
+        # if not a local file, get from model hub
+        if get_from_model_hub:
+            hf_model_name = "pytorch_model.bin"
+            revision = "main"
+
+            if "@" in model_name:
+                model_name_split = model_name.split("@")
+                revision = model_name_split[-1]
+                model_name = model_name_split[0]
+
+            # use model name as subfolder
+            if "/" in model_name:
+                model_folder = model_name.split("/", maxsplit=1)[1]
+            else:
+                model_folder = model_name
+
+            # Lazy import
+            from huggingface_hub import hf_hub_url, cached_download
+
+            url = hf_hub_url(model_name, revision=revision, filename=hf_model_name)
+
+            try:
+                model_path = cached_download(url=url, library_name="flair",
+                                             library_version=flair.__version__,
+                                             cache_dir=flair.cache_root / 'models' / model_folder)
+            except HTTPError as e:
+                # output information
+                log.error("-" * 80)
+                log.error(
+                    f"ACHTUNG: The key '{model_name}' was neither found on the ModelHub nor is this a valid path to a file on your system!")
+                # log.error(f" - Error message: {e}")
+                log.error(f" -> Please check https://huggingface.co/models?filter=flair for all available models.")
+                log.error(f" -> Alternatively, point to a model file on your local drive.")
+                log.error("-" * 80)
+                Path(flair.cache_root / 'models' / model_folder).rmdir()  # remove folder again if not valid
+
+        return model_path
 
     def get_transition_matrix(self):
         data = []
@@ -1107,6 +1018,10 @@ class SequenceTagger(flair.nn.Model):
                f'  (weights): {self.weight_dict}\n' + \
                f'  (weight_tensor) {self.loss_weights}\n)'
 
+    @property
+    def label_type(self):
+        return self.tag_type
+
 
 class MultiTagger:
     def __init__(self, name_to_tagger: Dict[str, SequenceTagger]):
@@ -1146,7 +1061,7 @@ class MultiTagger:
             tagger.predict(
                 sentences=sentences,
                 mini_batch_size=mini_batch_size,
-                all_tag_prob=all_tag_prob,
+                # all_tag_prob=all_tag_prob,
                 verbose=verbose,
                 label_name=name,
                 return_loss=return_loss,
