@@ -1,12 +1,12 @@
 import logging
-
 from abc import ABC, abstractmethod
-from typing import List, Callable, Tuple
+from typing import List, Callable, Optional
 
+from more_itertools import stagger
 from segtok.segmenter import split_single, split_multi
 from segtok.tokenizer import split_contractions, word_tokenizer
 
-from flair.data import Sentence, Tokenizer, Token
+from flair.data import Sentence, Token, Tokenizer
 
 log = logging.getLogger("flair")
 
@@ -56,7 +56,7 @@ class SpacyTokenizer(Tokenizer):
             tokens.append(token)
 
             if (previous_token is not None) and (
-                token.start_pos == previous_token.start_pos + len(previous_token.text)
+                    token.start_pos == previous_token.start_pos + len(previous_token.text)
             ):
                 previous_token.whitespace_after = False
 
@@ -67,11 +67,11 @@ class SpacyTokenizer(Tokenizer):
     @property
     def name(self) -> str:
         return (
-            self.__class__.__name__
-            + "_"
-            + self.model.meta["name"]
-            + "_"
-            + self.model.meta["version"]
+                self.__class__.__name__
+                + "_"
+                + self.model.meta["name"]
+                + "_"
+                + self.model.meta["version"]
         )
 
 
@@ -81,6 +81,7 @@ class SegtokTokenizer(Tokenizer):
 
         For further details see: https://github.com/fnl/segtok
     """
+
     def __init__(self):
         super(SegtokTokenizer, self).__init__()
 
@@ -134,6 +135,7 @@ class SpaceTokenizer(Tokenizer):
     """
         Tokenizer based on space character only.
     """
+
     def __init__(self):
         super(SpaceTokenizer, self).__init__()
 
@@ -178,7 +180,7 @@ class JapaneseTokenizer(Tokenizer):
             https://github.com/himkt/konoha
     """
 
-    def __init__(self, tokenizer: str, sudachi_mode: str="A"):
+    def __init__(self, tokenizer: str, sudachi_mode: str = "A"):
         super(JapaneseTokenizer, self).__init__()
 
         available_tokenizers = ["mecab", "janome", "sudachi"]
@@ -246,9 +248,9 @@ class JapaneseTokenizer(Tokenizer):
     @property
     def name(self) -> str:
         return (
-            self.__class__.__name__
-            + "_"
-            + self.tokenizer
+                self.__class__.__name__
+                + "_"
+                + self.tokenizer
         )
 
 
@@ -256,6 +258,7 @@ class TokenizerWrapper(Tokenizer):
     """
         Helper class to wrap tokenizer functions to the class-based tokenizer interface.
     """
+
     def __init__(self, tokenizer_func: Callable[[str], List[Token]]):
         super(TokenizerWrapper, self).__init__()
         self.tokenizer_func = tokenizer_func
@@ -336,7 +339,7 @@ class SciSpacyTokenizer(Tokenizer):
         infix_re = spacy.util.compile_infix_regex(infixes)
 
         self.model = spacy.load(
-            "en_core_sci_sm", disable=["tagger", "ner", "parser", "textcat"]
+            "en_core_sci_sm", disable=["tagger", "ner", "parser", "textcat", "lemmatizer"]
         )
         self.model.tokenizer.prefix_search = prefix_re.search
         self.model.tokenizer.infix_finditer = infix_re.finditer
@@ -356,7 +359,7 @@ class SciSpacyTokenizer(Tokenizer):
             tokens.append(token)
 
             if (previous_token is not None) and (
-                token.start_pos == previous_token.start_pos + len(previous_token.text)
+                    token.start_pos == previous_token.start_pos + len(previous_token.text)
             ) and (not word.text[0].isspace()):
                 previous_token.whitespace_after = False
 
@@ -367,11 +370,11 @@ class SciSpacyTokenizer(Tokenizer):
     @property
     def name(self) -> str:
         return (
-            self.__class__.__name__
-            + "_"
-            + self.model.meta["name"]
-            + "_"
-            + self.model.meta["version"]
+                self.__class__.__name__
+                + "_"
+                + self.model.meta["name"]
+                + "_"
+                + self.model.meta["version"]
         )
 
 
@@ -417,26 +420,31 @@ class SegtokSentenceSplitter(SentenceSplitter):
         self._tokenizer = tokenizer
 
     def split(self, text: str) -> List[Sentence]:
-        sentences = []
-        offset = 0
+        plain_sentences: List[str] = list(split_multi(text))
 
-        plain_sentences = split_multi(text)
-        for sentence in plain_sentences:
-            sentence_offset = text.find(sentence, offset)
+        try:
+            sentence_offset: Optional[int] = text.index(plain_sentences[0])
+        except ValueError as error:
+            raise AssertionError(f"Can't find the sentence offset for sentence {repr(plain_sentences[0])} "
+                                 f"from the text's starting position") from error
 
-            if sentence_offset == -1:
-                raise AssertionError(f"Can't find offset for sentences {plain_sentences} "
-                                     f"starting from {offset}")
+        sentences: List[Sentence] = []
+        for sentence, next_sentence in stagger(plain_sentences, offsets=(0, 1), longest=True):
 
-            sentences += [
+            sentences.append(
                 Sentence(
                     text=sentence,
                     use_tokenizer=self._tokenizer,
                     start_position=sentence_offset
                 )
-            ]
+            )
 
-            offset += len(sentence)
+            offset: int = sentence_offset + len(sentence)
+            try:
+                sentence_offset = text.index(next_sentence, offset) if next_sentence is not None else None
+            except ValueError as error:
+                raise AssertionError(f"Can't find the sentence offset for sentence {repr(sentence)} "
+                                     f"starting from position {repr(offset)}") from error
 
         return sentences
 
@@ -509,13 +517,13 @@ class SpacySentenceSplitter(SentenceSplitter):
     @property
     def name(self) -> str:
         return (
-            self.__class__.__name__
-            + "_"
-            + self.model.meta["name"]
-            + "_"
-            + self.model.meta["version"]
-            + "_"
-            + self._tokenizer.name
+                self.__class__.__name__
+                + "_"
+                + self.model.meta["name"]
+                + "_"
+                + self.model.meta["version"]
+                + "_"
+                + self._tokenizer.name
         )
 
 
@@ -573,11 +581,11 @@ class TagSentenceSplitter(SentenceSplitter):
     @property
     def name(self) -> str:
         return (
-            self.__class__.__name__
-            + "_"
-            + self.tag
-            + "_"
-            + self._tokenizer.name
+                self.__class__.__name__
+                + "_"
+                + self.tag
+                + "_"
+                + self._tokenizer.name
         )
 
 
@@ -593,9 +601,9 @@ class NewlineSentenceSplitter(TagSentenceSplitter):
     @property
     def name(self) -> str:
         return (
-            self.__class__.__name__
-            + "_"
-            + self._tokenizer.name
+                self.__class__.__name__
+                + "_"
+                + self._tokenizer.name
         )
 
 
@@ -628,7 +636,7 @@ class NoSentenceSplitter(SentenceSplitter):
     @property
     def name(self) -> str:
         return (
-            self.__class__.__name__
-            + "_"
-            + self._tokenizer.name
+                self.__class__.__name__
+                + "_"
+                + self._tokenizer.name
         )
